from datetime import datetime
from decimal import Decimal, ROUND_DOWN
from typing import List, Optional

import dateutil
from dateutil.parser import isoparse
from furl import furl

from .models import (
    IntervalConsumption,
    UnitType,
    Consumption,
    ElectricityMeter,
    MeterDirection,
    EnergyType,
    MeterGeneration,
    GasMeter,
    Tariff,
    Meter,
    Address,
    MeterPoint,
    PageReference,
    SortOrder,
    Aggregate,
    TariffRate,
)

_CUBIC_METERS_TO_KWH_MULTIPLIER = 11.1868
_KWH_TO_KWH_MULTIPLIER = 1
_UNIT_MULTIPLIERS = {
    (UnitType.KWH.value, UnitType.CUBIC_METERS.value): (1 / _CUBIC_METERS_TO_KWH_MULTIPLIER),
    (UnitType.CUBIC_METERS.value, UnitType.KWH.value): _CUBIC_METERS_TO_KWH_MULTIPLIER,
}
_QUANT_3DP = Decimal("0.001")


def to_timestamp_str(timestamp: datetime) -> str:
    """Convert a datetime to an iso timestamp string expected by the octopus energy APIs.

    Args:
        timestamp: The timestamp to convert.

    Returns:
        The timestamp in a iso format required by the octopus energy APIs
    """
    return timestamp.replace(microsecond=0).isoformat() if timestamp else None


def from_timestamp_str(timestamp: str) -> Optional[datetime]:
    """Convert am Octopus Energy timestamp string to an datetime.

    Args:
        timestamp: The timestamp to convert.

    Returns:
        The timestamp as a datetime object.
    """
    if timestamp is None:
        return None
    return dateutil.parser.isoparse(timestamp) if timestamp else None


def _map_tariffs(input_tariffs: List[dict]) -> List[Tariff]:
    return [
        Tariff(
            t["tariff_code"], from_timestamp_str(t["valid_from"]), from_timestamp_str(t["valid_to"])
        )
        for t in input_tariffs
    ]


def meters_from_response(response: dict) -> List[Meter]:
    meters = []
    for property_ in response["properties"]:
        address = Address(
            property_.get("address_line_1", None),
            property_.get("address_line_2", None),
            property_.get("address_line_3", None),
            property_.get("county", None),
            property_.get("town", None),
            property_.get("postcode", None),
            # the property is active if it has no moved out date
            property_.get("moved_out_at", None) is None,
        )
        for meter_point in property_.get("electricity_meter_points", []):
            for meter in meter_point["meters"]:
                meters.append(
                    ElectricityMeter(
                        meter_point=MeterPoint(meter_point["mpan"], address),
                        serial_number=meter["serial_number"],
                        tariffs=_map_tariffs(meter_point["agreements"]),
                        direction=MeterDirection.EXPORT
                        if meter_point["is_export"]
                        else MeterDirection.IMPORT,
                        energy_type=EnergyType.ELECTRICITY,
                        generation=MeterGeneration.SMETS1_ELECTRICITY,
                    )
                )
        for meter_point in property_.get("gas_meter_points", []):
            for meter in meter_point["meters"]:
                meters.append(
                    GasMeter(
                        meter_point=MeterPoint(meter_point["mprn"], address),
                        serial_number=meter["serial_number"],
                        tariffs=_map_tariffs(meter_point["agreements"]),
                        energy_type=EnergyType.GAS,
                        generation=MeterGeneration.SMETS1_GAS,
                    )
                )
        return meters


def consumption_from_response(
    response: dict, meter: Meter, desired_unit_type: UnitType = None
) -> Consumption:
    """Generates the Consumption model from an octopus energy API response.

    Args:
        response: The API response object.
        meter: The meter the consumption is related to.
        desired_unit_type: The desired unit for the consumption intervals. The mapping will
                           convert from the meters units to the desired units.

    Returns:
        The Consumption model for the period of time represented in the response.

    """
    if "results" not in response:
        return Consumption(unit_type=desired_unit_type, meter=meter)
    return Consumption(
        desired_unit_type,
        meter,
        [
            IntervalConsumption(
                consumed_units=_calculate_unit(
                    result["consumption"], meter.generation.unit_type, desired_unit_type
                ),
                interval_start=isoparse(result["interval_start"]),
                interval_end=isoparse(result["interval_end"]),
            )
            for result in response["results"]
        ],
        _get_page_reference(response, "previous"),
        _get_page_reference(response, "next"),
    )


def tariff_rates_from_response(response: dict) -> List[TariffRate]:
    """Generates the list of tariff rates from an octopus energy API response.

    Args:
        response: The API response object.

    Returns:
        The List containing the rates for a specific tariff

    """
    if "results" not in response:
        return []
    return [
        TariffRate(
            Decimal(result["value_exc_vat"]).quantize(_QUANT_3DP, rounding=ROUND_DOWN),
            Decimal(result["value_inc_vat"]).quantize(_QUANT_3DP, rounding=ROUND_DOWN),
            from_timestamp_str(result["valid_from"]),
            from_timestamp_str(result.get("valid_to", None)),
        )
        for result in response["results"]
    ]


def _get_page_reference(response: dict, page: str):
    if page not in response:
        return None
    page_url = furl(response[page])
    if not page_url.args:
        return None

    # Convert all args in the page reference to the types used by the APIs
    args = dict(page_url.args)
    if "period_from" in args:
        args["period_from"] = from_timestamp_str(args["period_from"])
    if "period_to" in args:
        args["period_to"] = from_timestamp_str(args["period_to"])
    if "order" in args:
        args["order"] = SortOrder(args["order"])
    if "group_by" in args:
        args["group_by"] = Aggregate(args["group_by"])

    return PageReference(args)


def _calculate_unit(consumption, actual_unit, desired_unit: UnitType = None):
    """Converts unit values from one unit to another unit.

    If no mapping is available the value is returned unchanged.

    :param consumption: The consumption to convert.
    :param actual_unit: The unit the supplied consumption is measured in.
    :param desired_unit: The unit the convert the consumption to.
    :return: The consumption converted to the desired unit.
    """
    return (
        consumption
        if not desired_unit
        else consumption * _UNIT_MULTIPLIERS.get((actual_unit.value, desired_unit.value), 1)
    )



"""Python client for the Octopus Energy RESTful API"""

from .models import (
    Address,
    Aggregate,
    Consumption,
    EnergyType,
    EnergyTariffType,
    IntervalConsumption,
    Meter,
    MeterDirection,
    MeterGeneration,
    MeterPoint,
    RateType,
    SortOrder,
    Tariff,
    UnitType,
    ElectricityMeter,
    GasMeter,
    PageReference,
    TariffRate,
)
from .exceptions import ApiAuthenticationError, ApiError, ApiNotFoundError, ApiBadRequestError
from .rest_client import OctopusEnergyRestClient
from .client import OctopusEnergyConsumerClient

__all__ = [
    "OctopusEnergyRestClient",
    "ApiAuthenticationError",
    "ApiError",
    "ApiNotFoundError",
    "ApiBadRequestError",
    "Consumption",
    "IntervalConsumption",
    "MeterGeneration",
    "UnitType",
    "Aggregate",
    "SortOrder",
    "RateType",
    "Meter",
    "EnergyType",
    "EnergyTariffType",
    "OctopusEnergyConsumerClient",
    "Tariff",
    "MeterDirection",
    "MeterPoint",
    "Address",
    "ElectricityMeter",
    "GasMeter",
    "PageReference",
    "TariffRate",
]



from datetime import datetime, timedelta
from typing import List, Optional

from .mappers import meters_from_response, consumption_from_response, tariff_rates_from_response
from octopus_energy import (
    Meter,
    OctopusEnergyRestClient,
    Consumption,
    EnergyType,
    SortOrder,
    PageReference,
    EnergyTariffType,
    RateType,
    TariffRate,
)


class OctopusEnergyConsumerClient:
    """An opinionated take on the consumer features of the Octopus Energy API.

    This client uses python model classes instead of raw json-as-a-dict. It does not expose all of
    the octopus energy API, only bits that are useful for a consumer of octopus energy products. It
    focuses on consumption and prices, and blending the two of them together to provide useful
    data for a consumer.

    This client is useful if you want to know your own personal energy consumption statistics.

    This client uses async i/o.
    """

    def __init__(self, api_token: Optional[str] = None):
        """Initializes the Octopus Energy Consumer Client.

        Args:
            api_token: Your Octopus Energy API Key.
        """
        self.rest_client = OctopusEnergyRestClient(api_token)

    def __enter__(self):
        raise TypeError("Use async context manager (async with) instead")

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        """Clean up resources used by the client.

        Once the client is closed, you cannot use it to make any further calls.
        """
        await self.rest_client.close()

    async def get_meters(self, account_number: str) -> List[Meter]:
        """Gets all meters associated with your account.

        Args:
            account_number: Your Octopus Energy Account Number.
        """
        return meters_from_response(await self.rest_client.get_account_details(account_number))

    async def get_consumption(
        self,
        meter: Meter,
        period_from: datetime = None,
        period_to: datetime = None,
        page_reference: PageReference = None,
    ) -> Consumption:
        """Get the energy consumption for a meter

        Args:
            meter: The meter to get consumption for.
            period_from: The timestamp for the earliest period of consumption to return.
            period_to: The timestamp for the latest period of consumption to return.
            page_reference: Get a specific page of results based on a page reference returned by
                            a previous call to get_consumption

        Returns:
            The consumption for the meter in the time period specified. The results are returned
            in ascending timestamp order from the start of the period.

        """
        func = (
            self.rest_client.get_electricity_consumption_v1
            if meter.energy_type == EnergyType.ELECTRICITY
            else self.rest_client.get_gas_consumption_v1
        )

        params = {}
        if page_reference:
            params.update(page_reference.options)
        else:
            params.update(
                {
                    "period_from": period_from,
                    "period_to": period_to,
                    "order": SortOrder.OLDEST_FIRST,
                }
            )

        response = await func(meter.meter_point.id, meter.serial_number, **params)
        return consumption_from_response(response, meter)

    async def get_tariff_cost(
        self,
        product_code: str,
        tariff_code: str,
        tariff_type: EnergyTariffType,
        rate_type: RateType,
        timestamp: datetime,
    ) -> Optional[TariffRate]:
        """Gets the cost of a tariff cost at a point in time.

        Args:
            product_code: The product code.
            tariff_code: The tariff code.
            tariff_type: The type of energy within the tariff.
            rate_type: The type of rate.
            timestamp: The timestamp

        Returns:
            The cost per unit of energy for the requested rate at a point in time.

        """
        response = await self.rest_client.get_tariff_v1(
            product_code,
            tariff_type,
            tariff_code,
            rate_type,
            period_from=timestamp,
            # Add a millisecond as the API doesn't like requests with the same start and end
            period_to=timestamp + timedelta(seconds=1),
        )
        rates = tariff_rates_from_response(response)
        return None if not rates else rates[0]



from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import List, Optional


class _DocEnum(Enum):
    """Wrapper to create enumerations with useful docstrings."""

    def __new__(cls, value, doc):
        self = object.__new__(cls)
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        return self


@dataclass
class PageReference:
    """Represents a reference to a page of information for API calls that support paging"""

    options: dict


@dataclass
class Tariff:
    code: str
    valid_from: datetime
    valid_to: Optional[datetime]


@dataclass
class TariffRate:
    cost_inc_vat: Decimal
    cost_exc_vat: Decimal
    valid_from: datetime
    valid_to: Optional[datetime]


class UnitType(Enum):
    """Units of energy measurement."""

    KWH = ("kWh", "Kilowatt Hours")
    CUBIC_METERS = ("m³", "Cubic Meters")

    @property
    def description(self) -> str:
        """A description, in english, of the unit type."""
        return self.value[1]

    def __eq__(self, other):
        return self.value == other.value


class MeterGeneration(Enum):
    """Energy meter types, the units the measure in and description in english."""

    SMETS1_GAS = ("SMETS1_GAS", UnitType.KWH, "1st Generation Smart Gas Meter")
    SMETS2_GAS = ("SMETS2_GAS", UnitType.CUBIC_METERS, "2nd Generation Smart Gas Meter")
    SMETS1_ELECTRICITY = (
        "SMETS1_ELECTRICITY",
        UnitType.KWH,
        "1st Generation Smart Electricity Meter",
    )
    SMETS2_ELECTRICITY = (
        "SMETS2_ELECTRICITY",
        UnitType.KWH,
        "2nd Generation Smart Electricity Meter",
    )

    @property
    def unit_type(self) -> UnitType:
        """The type of unit the meter measures consumption in."""
        return self.value[1]

    @property
    def description(self) -> str:
        """A description, in english, of the meter."""
        return self.value[2]


class EnergyType(Enum):
    """Represents a type of energy."""

    GAS = "gas"
    ELECTRICITY = "electricity"


class MeterDirection(_DocEnum):
    """Represents the direction that energy flows through the meter."""

    IMPORT = ("import", "Electricity that is consumed from the electricity grid")
    EXPORT = ("export", "Electricity that is sent to the electricity grid")


@dataclass
class Address:
    line_1: str
    line_2: str
    line_3: str
    county: str
    town: str
    postcode: str
    active: bool

    def __str__(self):
        """Gets a single line string representation of the address"""
        return (
            f"{self.line_1}"
            f"{', ' + self.line_2 if self.line_2 is not None else ''}"
            f"{', ' + self.line_3 if self.line_3 is not None else ''}"
            f"{', ' + self.county if self.county is not None else ''}"
            f"{', ' + self.town if self.town is not None else ''}"
            f"{', ' + self.postcode if self.postcode is not None else ''}"
        )


@dataclass
class MeterPoint:
    """Represents an energy meter point which has an identifier and is located at an address.

    Many meter points can share the same address.
    """

    id: str
    address: Address


@dataclass
class Meter:
    """Represents an energy meter, either gas or electric."""

    meter_point: MeterPoint
    serial_number: str
    energy_type: EnergyType
    generation: MeterGeneration
    tariffs: List[Tariff]

    def get_tariff_at(self, timestamp: datetime):
        """Gets the tariff in effect on a meter at a specific date/time.

        This automatically takes into account open ended tariffs that have no end."""
        return next(
            (
                tariff
                for tariff in self.tariffs
                if timestamp >= tariff.valid_from
                and (not tariff.valid_to or timestamp < tariff.valid_to)
            ),
            None,
        )


@dataclass
class ElectricityMeter(Meter):
    direction: MeterDirection


@dataclass
class GasMeter(Meter):
    pass


@dataclass
class IntervalConsumption:
    """Represents the consumption of energy over a single interval of time."""

    interval_start: datetime
    interval_end: datetime
    consumed_units: Decimal


@dataclass
class Consumption:
    """Consumption of energy for a list of time intervals."""

    unit_type: UnitType
    meter: Meter
    intervals: List[IntervalConsumption] = field(default_factory=lambda: [])
    previous_page: Optional[PageReference] = None
    next_page: Optional[PageReference] = None


class EnergyTariffType(_DocEnum):
    """Represents a type of energy tariff."""

    ELECTRICITY = (
        "electricity-tariffs",
        "Represents a type of tariff related to electricity consumption.",
    )
    GAS = ("gas-tariffs", "Represents a type of tariff related to gas consumption.")


class RateType(_DocEnum):
    """Represents a type of rate that can be charged for a product."""

    STANDING_CHARGES = (
        "standing-charges",
        "Represents the standing charge applied daily to an energy tariff.",
    )
    STANDARD_UNIT_RATES = (
        "standard-unit-rates",
        "Represents the standard unit rates for flat rate energy tariffs.",
    )
    DAY_UNIT_RATES = (
        "day-unit-rates",
        "Represents the rate charged during the day for dual rate tariffs.",
    )
    NIGHT_UNIT_RATES = (
        "night-unit-rates",
        "Represents the rate charged during the night for dual rate tariffs.",
    )


class SortOrder(_DocEnum):
    """The order to use when sorting results from search type API alls"""

    NEWEST_FIRST = (
        "-period",
        "Newest result first",
    )
    OLDEST_FIRST = (
        "period",
        "Oldest result first",
    )


class Aggregate(_DocEnum):
    """How to aggregate data returned from consumption based APIs"""

    HALF_HOURLY = None, "Aggregate consumption half hourly (the default)"
    HOUR = (
        "hour",
        "Aggregate consumption by hours",
    )
    DAY = (
        "day",
        "Aggregate consumption by days",
    )
    WEEK = (
        "week",
        "Aggregate consumption by weeks",
    )
    MONTH = (
        "month",
        "Aggregate consumption by months",
    )
    QUARTER = (
        "quarter",
        "Aggregate consumption quarterly",
    )
